/*  GNU Moe - My Own Editor
    Copyright (C) 2005-2016 Antonio Diaz Diaz.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <cctype>
#include <cstdio>
#include <ctime>
#include <string>
#include <vector>

#include "buffer.h"
#include "buffer_handle.h"
#include "block.h"
#include "encoding.h"
#include "iso_8859.h"
#include "menu.h"
#include "rc.h"
#include "regex.h"
#include "screen.h"
#include "window.h"
#include "window_vector.h"


namespace Window_vector {

int curwin_;			// current window (the one with the cursor in)
std::vector< Window > data;


const char * already_being_edited( const std::string & name,
                                   const Buffer & buffer )
  {
  if( name.size() )			// allow multiple unnamed buffers
    for( int i = 0; i < Bufhandle_vector::handles(); ++i )
      if( name == Bufhandle_vector::handle( i ).buffer().name() &&
          &buffer != &Bufhandle_vector::handle( i ).buffer() )
        {
        int bufsize = 0;
        char * const buf = Screen::line_buf( &bufsize );
        snprintf( buf, bufsize, "File '%s' is already being edited", name.c_str() );
        return buf;
        }
  return 0;
  }


bool no_block()
  {
  if( !Block::valid() ) { Screen::show_message( "No block" ); return true; }
  return false;
  }


bool no_block_in_this_file( const Buffer & buffer )
  {
  if( Block::bufferp() != &buffer )
    { Screen::show_message( "Block is not in this file" ); return true; }
  return false;
  }


bool rectangular_block()
  {
  if( RC::editor_options().rectangle_mode )
    {
    Screen::show_message( "Operation not available in rectangular block mode" );
    return true;
    }
  return false;
  }


bool read_only( const Buffer & buffer )
  {
  if( buffer.options.read_only )
    { Screen::show_message( "Read only" ); return true; }
  return false;
  }


void replace_and_repaint( Buffer & buffer, const Basic_buffer & tmp )
  {
  Point p = Block::end();
  buffer.reset_appendable();
  buffer.replace( Block::begin(), p, tmp, tmp.bof(), tmp.eof() );
  buffer.reset_appendable();
  if( RC::editor_options().auto_unmark ) Block::reset();
  else Block::set_end( buffer, p );
  Screen::repaint( &buffer );
  }


bool insert_year( std::vector< int > & years, const int year )
  {
  if( year >= 1000 )
    {
    unsigned i = years.size();
    while( i > 0 && years[i-1] > year ) --i;
    if( i == 0 || years[i-1] < year )
      { years.insert( years.begin() + i, year ); return true; }
    }
  return false;
  }

void print_year( Basic_buffer & outbuf, Point & p, int year )
  {
  while( year )
    { outbuf.pputc( p, '0' + ( year % 10 ) ); outbuf.pprev( p ); year /= 10; }
  p = outbuf.eof();
  }

bool insert_new_year( const Buffer & buffer, const Point & p0,
                      const Point & pn, const Point & py,
                      const Point & ph1, const Point & ph2,
                      const int new_year, Basic_buffer & outbuf )
  {
  std::vector< int > years;	// years are ordered and duplicates removed
  int year = 0, prev_year = 0;
  for( Point p = py; p <= ph1; buffer.pnext( p ) )	// parse years
    {
    const int ch = buffer[p];
    if( std::isdigit( ch ) ) { year *= 10; year += ch - '0'; continue; }
    if( ch < 0 || ch == '\n' || ch == ',' || ch == '-' || p >= ph1 )
      {
      if( prev_year >= 1000 )
        {
        insert_year( years, prev_year );
        for( int y = prev_year + 1; y < year; ++y ) insert_year( years, y );
        }
      insert_year( years, year );
      if( ch == '-' ) prev_year = year; else prev_year = 0;
      year = 0;
      if( ch < 0 ) break;
      }
    else if( std::ispunct( ch ) ) year = 0;
    }

  bool change = insert_year( years, new_year );		// new_year inserted

  const int rmargin = buffer.options.rmargin();
  Point p = outbuf.eof();
  int first_year = years[0];
  prev_year = first_year;
  for( unsigned i = 1; i <= years.size(); ++i )		// write years
    {
    year = ( i < years.size() ) ? years[i] : 0;
    if( prev_year + 1 != year )				// end of range
      {
      const bool range = first_year + 1 < prev_year;	// 3 or more years
      const int size = range ? 11 : 6;
      if( first_year > years[0] ) outbuf.pputc( p, ',' );
      if( outbuf.to_cursor( p ).col + size > rmargin + 1 )
        {
        outbuf.pputc( p, '\n' );
        outbuf.pputb( p, buffer, p0, pn );	// add prefix of notice
        }
      else if( first_year > years[0] ) outbuf.pputc( p, ' ' );
      if( first_year < prev_year )
        {
        print_year( outbuf, p, first_year );
        if( range ) outbuf.pputc( p, '-' );
        else { outbuf.pputc( p, ',' ); outbuf.pputc( p, ' ' ); }
        }
      print_year( outbuf, p, prev_year );
      first_year = year;
      }
    prev_year = year;
    }

  if( ph1 < ph2 )				// write copyright holder
    {
    Point pt = p;
    outbuf.pputb( p, buffer, ph1, ph2 );
    if( outbuf.to_cursor( p ).col + 1 > rmargin + 1 )
      {
      outbuf.pputc( pt, '\n' );
      outbuf.pputb( pt, buffer, p0, pn );	// add prefix of notice
      }
    else outbuf.pputc( pt, ' ' );
    }

  if( !change ) change = ( outbuf.last_line() != ( ph2.line - p0.line ) );
  if( !change )
    {
    std::string s1, s2;
    buffer.to_string( p0, ph2, s1 );
    outbuf.to_string( outbuf.bof(), outbuf.eof(), s2 );
    change = ( s1 != s2 );
    }
  return change;			// new_year inserted or list modified
  }

} // end namespace Window_vector


int Window_vector::init()
  {
  curwin_ = 0;
  return load( RC::editor_options().orphan_extra ? 1 : Bufhandle_vector::handles() );
  }


// 'new_size' == 0: reload same number of windows
// 'new_size' < 0: toggle 1 window <--> max_windows
//
int Window_vector::load( int new_size )
  {
  const int handles = Bufhandle_vector::handles();
  const bool toggle = ( new_size < 0 && handles > 1 );
  const bool center = data.empty();			// center all on init

  if( !handles ) { curwin_ = 0; data.clear(); return 0; }
  if( new_size == 0 ) new_size = data.size();
  else if( new_size < 0 )
    {
    for( int i = 0; i < curwin_; ++i ) Bufhandle_vector::next();
    new_size = ( data.size() != 1 ) ? 1 : handles;
    curwin_ = 0;
    }
  new_size = std::min( new_size, handles );
  new_size = std::min( new_size, Screen::max_windows() );
  if( RC::editor_options().max_windows() > 0 )
    new_size = std::min( new_size, RC::editor_options().max_windows() );
  new_size = std::max( new_size, 1 );
  while( curwin_ >= new_size ) { --curwin_; Bufhandle_vector::next(); }
  const int wheight = Screen::height() / new_size;

  data.clear();
  int i = 0, y = 0;
  while( i + 1 < new_size )
    {
    data.push_back( Window( Bufhandle_vector::handle( i ), y, wheight, center ) );
    ++i; y += wheight;
    }
  data.push_back( Window( Bufhandle_vector::handle( i ), y, Screen::height() - y, center ) );
  if( toggle && data.size() == 1 ) data[0].center_cursor();
  for( i = 0; i < (int)data.size(); ++i ) data[i].repaint();
  return data.size();
  }


void Window_vector::bufhandle_menu( const int abort_key )
  {
  const int handle = Menu::bufhandle_menu( curwin_, abort_key );
  if( handle >= 0 ) { curwin_ = handle; if( curwin_ >= windows() ) load(); }
  }


const char * Window_vector::change_buffer_name()
  {
  Buffer & buffer = data[curwin_].buffer();
  std::string name( buffer.name() );

  if( Menu::get_filename( "New buffer name (^C to abort): ", name ) < 0 ||
      name == buffer.name() )
    return 0;
  const char * const msg = already_being_edited( name, buffer );
  if( !msg )
    {
    buffer.name( name );
    Screen::show_status_lines( &buffer );
    }
  return msg;
  }


const char * Window_vector::load_file( const int abort_key )
  {
  std::string name;
  if( Menu::get_filename( "Name of file to edit (^C to abort): ",
                          name, abort_key ) < 0 )
    return 0;
  const int old_handles = Bufhandle_vector::handles();
  int handle;
  try { handle = Bufhandle_vector::find_or_add_handle( &name, curwin_ + 1 ); }
  catch( Buffer::Error e ) { return e.msg; }
  curwin_ = handle;
  if( windows() > 1 && Bufhandle_vector::handles() > old_handles )
    load( windows() + 1 );
  else load();
  return 0;
  }


void Window_vector::split_window()
  {
  Bufhandle_vector::duplicate_handle( curwin_++ );
  load( windows() + 1 );
  }


const char * Window_vector::save_file( const int abort_key )
  {
  const Buffer & buffer = data[curwin_].buffer();
  std::string name( buffer.name() );

  if( Menu::get_filename( "Name of file to save (^C to abort): ",
      name, abort_key ) <= 0 )
    return 0;
  const char * const msg = already_being_edited( name, buffer );
  if( msg ) return msg;
  bool done;
  try { done = buffer.save( &name ); }
  catch( Buffer::Error e ) { return e.msg; }
  if( done )
    {
    Screen::show_status_lines( &buffer );
    int bufsize = 0;
    char * const buf = Screen::line_buf( &bufsize );
    snprintf( buf, bufsize, "File '%s' saved", name.c_str() );
    return buf;
    }
  if( !read_only( buffer ) ) return "File not saved";
  return 0;
  }


// Returns number of windows, 0 if none, -1 if aborted.
//
int Window_vector::close( bool abort, const bool quiet )
  {
  Buffer_handle & bh = data[curwin_].buffer_handle();
  const bool last_handle = Bufhandle_vector::handles( bh.buffer() ) <= 1;
  const bool modified = bh.buffer().modified();
  const bool exit_ask = RC::editor_options().exit_ask && !quiet;

  if( !abort || !quiet )
    {
    if( abort && modified && last_handle )
      {
      int key = Screen::show_message( "Lose changes to this file ? (y/n/^C) ", true, true );
      if( key == 'n' || key == 'N' ) abort = false;
      else if( key != 'y' && key != 'Y' ) return -1;
      }
    if( !abort && modified )
      {
      if( read_only( bh.buffer() ) ) return -1;
      std::string name( bh.buffer().name() );
      if( exit_ask || name.empty() )
        while( !Menu::get_filename( "Name of file to save (^C to abort): ", name ) ) ;
      if( name.empty() ) return -1;
      try { if( !bh.buffer().save( &name ) ) return -1; }
      catch( Buffer::Error e )
        { if( !quiet ) Screen::show_message( e.msg ); return -1; }
      }
    else if( exit_ask && last_handle && !modified )
      {
      int key = Screen::show_message( "Abandon this file ? (y/N) ", true, true );
      if( key != 'y' && key != 'Y' ) return -1;
      }
    }
  Bufhandle_vector::delete_handle( curwin_ );
  if( curwin_ > 0 ) --curwin_;
  return load();
  }


int Window_vector::close_and_exit( const bool abort )
  {
  int key;
  if( abort ) key = Screen::show_message( "Are you sure you want to abort ? (y/N) ", true, true );
  else key = Screen::show_message( "Are you sure you want to exit ? (y/N) ", true, true );
  if( key == 'y' || key == 'Y' )
    {
    prev();
    while( data.size() && close( abort, true ) >= 0 ) ;
    }
  return data.size();
  }


Window & Window_vector::curwin() { return data[curwin_]; }


void Window_vector::last_visited()
  {
  if( Bufhandle_vector::handles() <= 1 ) return;
  curwin_ = Bufhandle_vector::last_visited( curwin_ );
  if( curwin_ >= windows() ) load();
  }


void Window_vector::next()
  {
  const int handles = Bufhandle_vector::handles();
  if( handles <= 1 ) return;
  if( curwin_ + 1 < windows() ) ++curwin_;
  else
    {
    if( handles <= windows() ) curwin_ = 0;
    else { Bufhandle_vector::next(); load(); }
    }
  }


void Window_vector::prev()
  {
  const int handles = Bufhandle_vector::handles();
  if( handles <= 1 ) return;
  if( curwin_ > 0 ) --curwin_;
  else
    {
    if( handles <= windows() ) curwin_ = windows() - 1;
    else { Bufhandle_vector::prev(); load(); }
    }
  }


Window & Window_vector::window( int i )
  {
  if( i < 0 ) i = 0; else if( i >= windows() ) i = windows() - 1;
  return data[i];
  }


int Window_vector::windows() { return data.size(); }


const char * Window_vector::add_char( const int key, const bool force )
  {
  Window & w = data[curwin_];
  if( w.buffer().options.read_only )
    {
    if( !force ) switch( key )
      {
      case ' ': w.move_vertical( w.height() - 2 ); return 0;
      case '/': return search( 6 );				// ^F
      case 'n': return search( 7, true );			// ^G
      }
    read_only( w.buffer() );
    return 0;
    }

  if( key >= 0 && key < 256 &&
      ( force || key == '\t' || key == '\n' || ISO_8859::isprint( key ) ) )
    { if( w.buffer().pputc( w.pointer(), key ) )
        Screen::repaint( &w.buffer() ); }
  else
    {
    int bufsize = 0;
    char * const buf = Screen::line_buf( &bufsize );
    snprintf( buf, bufsize, "Key code %d(0x%02X) ignored", key, key );
    return buf;
    }
  return 0;
  }


void Window_vector::delete_char( const bool back )
  {
  Window & w = data[curwin_];
  Point & p = w.pointer();

  if( !read_only( w.buffer() ) && w.buffer().pdelc( p, back ) )
    Screen::repaint( &w.buffer() );
  }


void Window_vector::delete_line()
  {
  Window & w = data[curwin_];

  if( !read_only( w.buffer() ) && w.buffer().pdell( w.pointer() ) )
    Screen::repaint( &w.buffer() );
  }


void Window_vector::copy_block()
  {
  Window & w = data[curwin_];

  if( !no_block() && !read_only( w.buffer() ) )
    {
    const Buffer & buffer = *Block::bufferp();
    if( Block::copy_block( w.buffer(), w.pointer() ) )
      {
      if( &buffer != &w.buffer() ) Screen::repaint( &buffer );
      Screen::repaint( &w.buffer() );
      }
    }
  }


void Window_vector::delete_block()
  {
  Buffer & buffer = data[curwin_].buffer();

  if( !no_block() && !no_block_in_this_file( buffer ) &&
      !read_only( buffer ) && Block::delete_block() )
    Screen::repaint( &buffer );
  }


void Window_vector::move_block()
  {
  Window & w = data[curwin_];
  Point & p = w.pointer();

  if( !no_block() && !Block::in_block_or_end( w.buffer(), p ) )
    {
    const Buffer & buffer = *Block::bufferp();
    if( !read_only( w.buffer() ) && !read_only( buffer ) &&
        Block::copy_block( w.buffer(), p, true ) )
      {
      if( &buffer != &w.buffer() ) Screen::repaint( &buffer );
      Screen::repaint( &w.buffer() );
      }
    }
  }


const char * Window_vector::read_block()
  {
  Window & w = data[curwin_];
  if( read_only( w.buffer() ) ) return 0;

  std::string name;
  if( Menu::get_filename( "Name of file to insert (^C to abort): ", name ) <= 0 )
    return 0;
  try {
    Buffer tmp( &name );
    if( tmp.empty() ) throw Buffer::Error( "File not found" );
    Point p = w.pointer();
    if( w.buffer().pputb( p, tmp, tmp.bof(), tmp.eof() ) )
      Screen::repaint( &w.buffer() );
    }
  catch( Buffer::Error e ) { return e.msg; }
  return 0;
  }


const char * Window_vector::write_block()
  {
  if( no_block() ) return 0;
  std::string name;
  if( Menu::get_filename( "Name of file to write (^C to abort): ", name ) <= 0 )
    return 0;
  try {
    if( !RC::editor_options().rectangle_mode )
      Block::bufferp()->save( &name, false, false, Block::begin(), Block::end() );
    else
      {
      Buffer tmp;
      Block::save_block_position();
      Block::copy_block( tmp, tmp.bof() );
      Block::restore_block_position();
      tmp.save( &name );
      }
    if( RC::editor_options().auto_unmark )
      {
      const Buffer * const bufferp = Block::bufferp();
      Block::reset(); Screen::repaint( bufferp );
      }
    }
  catch( Buffer::Error e ) { return e.msg; }
  return 0;
  }


void Window_vector::indent_block()
  {
  Buffer & buffer = data[curwin_].buffer();

  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return;
  Point p = Block::begin();
  bool first_post = true;
  const bool ow = buffer.options.overwrite;
  buffer.options.overwrite = false;
  buffer.reset_appendable();
  while( buffer.bol( p ) < Block::end() )
    {
    p = buffer.bot( p );
    if( p.col > 0 || p < buffer.eol( p ) )
      {
      if( first_post ) first_post = false; else buffer.force_append();
      for( int i = 0; i < RC::editor_options().indent_step(); ++i )
        buffer.pputc( p, ' ' );
      }
    ++p.line;
    }
  buffer.reset_appendable();
  buffer.options.overwrite = ow;
  if( RC::editor_options().auto_unmark ) Block::reset();
  Screen::repaint( &buffer );
  }


const char * Window_vector::unindent_block()
  {
  Buffer & buffer = data[curwin_].buffer();

  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return 0;
  Point p = Block::begin();
  bool fail = false;
  while( buffer.bol( p ) < Block::end() && !fail )
    {
    if( buffer.eol( p ).col > 0 &&
        buffer.to_cursor( buffer.bot( p ) ).col < RC::editor_options().indent_step() )
      fail = true;
    ++p.line;
    }
  if( fail ) return "Can't unindent more";

  bool first_post = true;
  const bool ow = buffer.options.overwrite;
  buffer.options.overwrite = false;
  buffer.reset_appendable();
  p = Block::begin();
  while( buffer.bol( p ) < Block::end() )
    {
    const int target_col = buffer.to_cursor( buffer.bot( p ) ).col -
                           RC::editor_options().indent_step();
    if( buffer.eol( p ).col > 0 )
      while( true )
        {
        p = buffer.bot( p );
        const int col = buffer.to_cursor( p ).col;
        if( col > target_col )
          {
          if( first_post ) first_post = false; else buffer.force_append();
          buffer.pdelc( p, true );
          }
        else if( col < target_col )
          {
          if( first_post ) first_post = false; else buffer.force_append();
          buffer.pputc( p, ' ' );
          }
        else break;
        }
    ++p.line;
    }
  buffer.reset_appendable();
  buffer.options.overwrite = ow;
  if( RC::editor_options().auto_unmark ) Block::reset();
  Screen::repaint( &buffer );
  return 0;
  }


const char * Window_vector::undo()
  {
  Window & w = data[curwin_];

  if( !read_only( w.buffer() ) )
    {
    if( w.buffer().undo( w.pointer() ) ) Screen::repaint( &w.buffer() );
    else return "Nothing to undo";
    }
  return 0;
  }


const char * Window_vector::redo()
  {
  Window & w = data[curwin_];

  if( !read_only( w.buffer() ) )
    {
    if( w.buffer().redo( w.pointer() ) ) Screen::repaint( &w.buffer() );
    else return "Nothing to redo";
    }
  return 0;
  }


void Window_vector::center_line()
  {
  Window & w = data[curwin_];
  Buffer & buffer = w.buffer();
  const Point bol = buffer.bol( w.pointer() );
  const Point bot = buffer.bot( bol );
  const Point cbot = buffer.to_cursor( bot );
  const Point ceot = buffer.to_cursor( buffer.eot( bol ) );
  int l = cbot.col - buffer.options.lmargin();
  int r = buffer.options.rmargin() - ceot.col;

  if( !read_only( buffer ) && cbot < ceot && l != r && l != r + 1 )
    {
    buffer.reset_appendable();
    const int off = ( bot != cbot ) ? cbot.col : 0;
    if( off )
      { buffer.pdelb( bol, bot ); l -= off; r += off; buffer.force_append(); }
    else if( l > r + 1 && bot.col > 0 )
      buffer.pdelb( bol, Point( bot.line, std::min( bot.col, ( l - r ) / 2 ) ) );
    if( l < r )
      {
      const bool ow = buffer.options.overwrite;
      buffer.options.overwrite = false;
      Point p = bol;
      for( ; l < r; ++l, --r ) buffer.pputc( p, ' ' );
      buffer.options.overwrite = ow;
      }
    buffer.reset_appendable();
    Screen::repaint( &buffer );
    }
  }


const char * Window_vector::reformat()
  {
  Window & w = data[curwin_];
  Buffer_handle & bh = w.buffer_handle();
  Buffer & buffer = bh.buffer();
  const Point bop = buffer.bop( bh.pointer );	// begin of paragraph
  const Point eop = buffer.eop( bh.pointer );	// end of paragraph
  const Point bot = buffer.bot( bop );		// begin of text in line

  if( !read_only( buffer ) && bop < eop )
    {
    const int rmargin = buffer.options.rmargin();
    const int lmargin = std::max( buffer.options.lmargin(),
                                  std::min( rmargin - 1, bh.cursor.col ) );
    Buffer tmp;
    tmp.options.set_rmargin( rmargin );
    tmp.options.auto_indent = false;
    tmp.options.overwrite = false;
    tmp.options.word_wrap = true;
    Point dummy = tmp.bof();
    if( bop < bot && buffer.to_cursor( bot ).col > lmargin )
      tmp.Basic_buffer::pputb( dummy, buffer, bop, bot );
    tmp.options.set_lmargin( lmargin );
    bool space = false;
    for( Point p = bot; p < eop; )
      {
      unsigned char ch = buffer.pgetc( p );
      if( !std::isspace( ch ) )			// exclude no-break space
        { tmp.pputc( dummy, ch ); dummy = tmp.eof(); space = false; }
      else if( !space ) { tmp.pputc( dummy, ' ' ); space = true; }
      if( Screen::kbhit() >= 0 && Screen::wait_kbhit() == 3 )
        return "Aborted";
      }
    if( space ) tmp.pdelc( dummy, true );
    if( dummy != tmp.bol( dummy ) ) tmp.pputc( dummy, '\n' );
    bool change = ( tmp.last_line() != ( eop.line - bop.line ) );
    if( !change )
      for( int line = 0; line < tmp.lines(); ++line )
        if( tmp.characters( line ) != buffer.characters( bop.line + line ) &&
            ( !tmp.blank( line ) || !buffer.blank( bop.line + line ) ) )
          { change = true; break; }
    if( change )
      {
      buffer.reset_appendable();
      Point p = eop;
      buffer.replace( bop, p, tmp, tmp.bof(), tmp.eof() );
      buffer.reset_appendable();
      Screen::repaint( &buffer );
      p.line = bop.line; p.col = lmargin;
      w.update_points( p, true, false, Window::from_cursor );
      }
    }
  return 0;
  }


const char * Window_vector::change_case( const bool to_caps )
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || read_only( buffer ) )
    return 0;
  Basic_buffer tmp;
  Point p = tmp.bof();
  bool change = false;
  if( to_caps )
    for( Point p1 = Block::begin(); p1 < Block::end(); )
      {
      const bool in_block = Block::in_block( buffer, p1 );
      unsigned char ch = buffer.pgetc( p1 );
      if( in_block && ISO_8859::islower( ch ) )
        { ch = ISO_8859::toupper( ch ); change = true; }
      tmp.pputc( p, ch );
      }
  else
    for( Point p1 = Block::begin(); p1 < Block::end(); )
      {
      const bool in_block = Block::in_block( buffer, p1 );
      unsigned char ch = buffer.pgetc( p1 );
      if( in_block && ISO_8859::isupper( ch ) )
        { ch = ISO_8859::tolower( ch ); change = true; }
      tmp.pputc( p, ch );
      }
  if( !change ) return "No letters to change in block";
  replace_and_repaint( buffer, tmp );
  return 0;
  }


void Window_vector::encode_base64()
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return;
  std::string in, out;
  Block::bufferp()->to_string( Block::begin(), Block::end(), in );
  Encoding::base64_encode( in, out );
  Basic_buffer tmp;
  Point p = tmp.bof();
  int c = 0;
  for( unsigned i = 0; i < out.size(); ++i )
    {
    tmp.pputc( p, out[i] );
    if( ++c >= 72 ) { c = 0; tmp.pputc( p, '\n' ); }
    }
  if( c ) tmp.pputc( p, '\n' );
  replace_and_repaint( buffer, tmp );
  }


const char * Window_vector::decode_base64()
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return 0;
  std::string in, out;
  Block::bufferp()->to_string( Block::begin(), Block::end(), in );
  for( int i = in.size() - 1; i >= 0; --i )
    if( ISO_8859::isspace( in[i] ) )
      in.erase( i, 1 );
  int idx;
  if( !Encoding::base64_decode( in, out, &idx ) )
    {
    Block::disable_marking();
    Point p = Block::begin();
    while( idx > 0 ) if( !ISO_8859::isspace( buffer.pgetc( p ) ) ) --idx;
    while( ISO_8859::isspace( buffer[p] ) ) buffer.pnext( p );
    if( idx == 0 && buffer.pisvalid( p ) )
      data[curwin_].update_points( p, true, true );
    return "Invalid Base64 data in block";
    }
  Basic_buffer tmp;
  Point p = tmp.bof();
  for( unsigned i = 0; i < out.size(); ++i )
    tmp.pputc( p, out[i] );
  replace_and_repaint( buffer, tmp );
  return 0;
  }


bool Window_vector::encode_ascii_utf8( const char ** const msgp,
                                       const bool mode_a, const bool force )
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) { *msgp = 0; return false; }
  std::string in, out;
  Block::bufferp()->to_string( Block::begin(), Block::end(), in );
  if( mode_a && !Encoding::ascii_encode( in, out ) )
    { *msgp = "Text in block is already ASCII encoded"; return false; }
  if( !mode_a && !Encoding::utf8_encode( in, out, force ) )
    { *msgp = "Text in block is already UTF-8 encoded"; return false; }
  Basic_buffer tmp;
  Point p = tmp.bof();
  for( unsigned i = 0; i < out.size(); ++i )
    tmp.pputc( p, out[i] );
  replace_and_repaint( buffer, tmp );
  *msgp = "Done";
  return true;
  }


const char * Window_vector::decode_quoted_printable_utf8( const bool mode_q,
                                                          const bool force )
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return 0;
  std::string in, out;
  Block::bufferp()->to_string( Block::begin(), Block::end(), in );
  if( mode_q )
    {
    int idx;
    if( !Encoding::quoted_printable_decode( in, out, &idx ) )
      {
      Block::disable_marking();
      Point p = Block::begin();
      if( buffer.pisvalid( p ) && buffer.pseek( p, idx ) )
        data[curwin_].update_points( p, true, true );
      return "Invalid Quoted-Printable data in block";
      }
    }
  else
    {
    int idx;
    const int tmp = Encoding::utf8_decode( in, out, &idx, force );
    if( tmp != 0 )
      {
      Block::disable_marking();
      Point p = Block::begin();
      if( buffer.pisvalid( p ) && buffer.pseek( p, idx ) )
        data[curwin_].update_points( p, true, true );
      if( tmp > 0 )
        {
        int bufsize = 0;
        char * const buf = Screen::line_buf( &bufsize );
        snprintf( buf, bufsize, "UTF-8 code 'U%04X' out of range in block", tmp );
        return buf;
        }
      else
        return "Invalid UTF-8 data in block";
      }
    }
  if( in == out ) return "Text in block is already decoded";
  Basic_buffer tmp;
  Point p = tmp.bof();
  for( unsigned i = 0; i < out.size(); ++i )
    tmp.pputc( p, out[i] );
  replace_and_repaint( buffer, tmp );
  return "Done";
  }


const char * Window_vector::remove_utf8_out_of_range()
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return 0;
  std::string in, out;
  Block::bufferp()->to_string( Block::begin(), Block::end(), in );
  int idx;
  const int ret = Encoding::remove_utf8_out_of_range( in, out, &idx );
  if( ret == 0 ) return "Text in block is already decoded";
  if( ret < 0 )
    {
    Block::disable_marking();
    Point p = Block::begin();
    if( buffer.pisvalid( p ) && buffer.pseek( p, idx ) )
      data[curwin_].update_points( p, true, true );
    return "Invalid UTF-8 data in block";
    }
  Basic_buffer tmp;
  Point p = tmp.bof();
  for( unsigned i = 0; i < out.size(); ++i )
    tmp.pputc( p, out[i] );
  replace_and_repaint( buffer, tmp );
  return "Done";
  }


void Window_vector::encode_rot1347( const bool mode13 )
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return;
  Basic_buffer tmp;
  Point p = tmp.bof();
  if( mode13 )
    for( Point p1 = Block::begin(); p1 < Block::end(); )
      tmp.pputc( p, Encoding::rot13( buffer.pgetc( p1 ) ) );
  else
    for( Point p1 = Block::begin(); p1 < Block::end(); )
      tmp.pputc( p, Encoding::rot47( buffer.pgetc( p1 ) ) );
  replace_and_repaint( buffer, tmp );
  }


const char * Window_vector::remove_duplicate_lines( const bool back )
  {
  Buffer & buffer = data[curwin_].buffer();
  if( no_block() || no_block_in_this_file( buffer ) || rectangular_block() ||
      read_only( buffer ) ) return 0;

  bool change = false;
  if( back )
    {
    int first_line = Block::begin().line;
    if( Block::begin().col > 0 ) ++first_line;
    for( int line = Block::end().line - 1; line > first_line; --line )
      for( Point p( line - 1, 0 ); p.line >= first_line; --p.line )
        if( buffer.compare_lines( line, p.line ) )
          {
          if( change ) buffer.force_append(); else change = true;
          buffer.pdell( p );
          --line;
          }
    }
  else
    {
    int line = Block::begin().line;
    if( Block::begin().col > 0 ) ++line;
    for( ; line < Block::end().line - 1; ++line )
      for( Point p( Block::end().line - 1, 0 ); p.line > line; --p.line )
        if( buffer.compare_lines( line, p.line ) )
          {
          if( change ) buffer.force_append(); else change = true;
          buffer.pdell( p );
          }
    }
  if( !change ) return "No duplicate lines in block";
  buffer.reset_appendable();
  if( Block::valid() )
    {
    if( back ) data[curwin_].update_points( Block::end(), true, true );
    else data[curwin_].update_points( Block::begin(), true, true );
    }
  if( RC::editor_options().auto_unmark ) Block::reset();
  Screen::repaint( &buffer );
  return 0;
  }


const char * Window_vector::copyright_update()
  {
  static std::vector< std::string > history;
  const std::string search_string( "\\<Copyright\\>\\+\\[^\\n]\\[1-9]\\+\\+\\+\\+\\[0-9]\\*\\<\\[^0-9]" );
  const std::string search_string2( "\\[1-9]\\+\\+\\+\\+\\[0-9]" );

  std::string year_string;
  const time_t tt = std::time( 0 );
  const std::tm * const t = std::localtime( &tt );
  if( t && t->tm_year >= 0 )
    {
    int year = 1900 + t->tm_year;
    while( year )
      { year_string.insert( 0U, 1, '0' + ( year % 10 ) ); year /= 10; }
    }
  history.push_back( year_string );
  if( Screen::get_string( "Year to add (^C to abort): ", history ) <= 0 )
    return 0;
  year_string = history.back();
  int new_year = 0;
  for( unsigned i = 0; i < year_string.size(); ++i )
    {
    if( !std::isdigit( year_string[i] ) ) { year_string.clear(); break; }
    new_year *= 10; new_year += year_string[i] - '0';
    }
  if( year_string.size() < 4 || year_string[0] == '0' ||
      new_year < 1000 || new_year > 9999 ) return "Invalid year";
  bool ask = true, found = false, modified = false, outdated = false;
  bool skipped = false;
  for( int i = 0; i < Bufhandle_vector::handles(); next(), ++i )
    {
    Window & w = data[curwin_];
    if( w.buffer().options.read_only ) { skipped = true; continue; }
    Point p = w.buffer().bof();
    std::vector< std::string > regex_pieces;
    while( true )
      {
      Point pn = p;			// begin of notice
      if( Regex::find( w.buffer(), pn, p, search_string, regex_pieces ) != 0 )
        break;				// no more copyrights found
      Point py = pn;			// begin of range of years
      Point pt;
      if( Regex::find( w.buffer(), py, pt, search_string2, regex_pieces ) != 0 ||
          py <= pn || pt >= p )
        continue;			// not a valid copyright notice
      found = true;
      const Point p0 = w.buffer().bol( pn );	// bol of notice
      const Point ph2 = w.buffer().eol( p );	// end of copyright holder
      w.buffer().pprev( p );			// begin of copyright holder
      Basic_buffer tmp( w.buffer(), p0, py );	// "<prefix> Copyright <text>"
      if( insert_new_year( w.buffer(), p0, pn, py, p, ph2, new_year, tmp ) )
        {
        outdated = true;
        p = ph2;
        w.update_points( p, true, true );
        int key = 0;
        if( ask || Screen::kbhit() >= 0 )
          {
          key = Screen::show_message( "Update (Y)es (N)o (R)est (^C to abort): ", true );
          if( key == 3 ) return 0;
          if( ask && ( key == 'r' || key == 'R' ) ) ask = false;
          }
        if( !ask || key == 'y' || key == 'Y' )
          {
          w.buffer().replace( p0, p, tmp, tmp.bof(), tmp.eof() );
          w.update_points( p, true, true );
          Screen::repaint( &w.buffer() );
          modified = true;
          }
        }
      }
    }
  if( skipped )
    {
    if( !found ) return "No copyright notices found (read-only buffers were skipped)";
    if( !outdated ) return "All copyright notices are up to date (read-only buffers were skipped)";
    if( !modified ) return "No copyright notices have been modified (read-only buffers were skipped)";
    return "Finished (read-only buffers were skipped)";
    }
  else
    {
    if( !found ) return "No copyright notices found";
    if( !outdated ) return "All copyright notices are up to date";
    if( !modified ) return "No copyright notices have been modified";
    return "Finished";
    }
  }


const char * Window_vector::search_word( const int abort_key )
  {
  const Buffer & buffer = data[curwin_].buffer();
  Point p1 = data[curwin_].pointer();
  if( ISO_8859::isalnum_( buffer[p1] ) || buffer.piseow( p1 ) )
    {
    Point p2 = p1;
    while( !buffer.pisbow( p1 ) ) if( !buffer.pprev( p1 ) ) break;
    while( !buffer.piseow( p2 ) ) if( !buffer.pnext( p2 ) ) break;
    if( buffer.pisbow( p1 ) && buffer.piseow( p2 ) )
      {
      std::string s_ini;
      buffer.to_string( p1, p2, s_ini );
      s_ini.insert( 0, "\\<" ); s_ini += "\\>";
      return search( abort_key, false, s_ini );
      }
    }
  return 0;
  }


const char * Window_vector::search( const int abort_key, const bool again,
                                    const std::string & s_ini,
                                    const bool toggle )
  {
  static std::vector< std::string > search_history;
  static std::vector< std::string > replace_history;
  static std::vector< std::string > options_history;
  static std::string search_string;
  static std::string replace_string;
  static bool backward = false;
  static bool block = false;
  static bool global = false;
  static bool icase = false;
  static bool replace = false;

  if( !again || search_string.empty() )
    {
    search_history.push_back( s_ini );
    int size = Screen::get_string( "Find (^C to abort): ", search_history,
                                   abort_key );
    if( size <= 0 )
      {
      if( s_ini.size() && search_history.size() ) search_history.pop_back();
      if( search_history.empty() ) search_string.clear();
      return 0;
      }
    search_string = search_history.back();
    ISO_8859::deescapize_search_string( search_string );
    std::string prompt;
    if( RC::editor_options().ignore_case )
      prompt = "(N)o_ignore_case (R)eplace (B)ackwards (G)lobal Bloc(K) (^C to abort): ";
    else prompt = "(I)gnore_case (R)eplace (B)ackwards (G)lobal Bloc(K) (^C to abort): ";
    options_history.push_back( std::string() );
    size = Screen::get_string( prompt, options_history, abort_key );
    if( size < 0 ) return 0;
    backward = false; block = false; global = false; replace = false;
    icase = RC::editor_options().ignore_case;
    if( size > 0 )
      {
      const std::string & options = options_history.back();
      for( unsigned i = 0; i < options.size(); ++i )
        switch( ISO_8859::toupper( options[i] ) )
          {
          case 'B': backward = true; break;
          case 'G': global = true; break;
          case 'I': icase = true; break;
          case 'K': block = true; break;
          case 'N': icase = false; break;
          case 'R': replace = true; break;
          }
      }
    if( block ) global = false;
    if( replace )
      {
      replace_string.clear();
      replace_history.push_back( std::string() );
      const int len = Screen::get_string( "Replace with (^C to abort): ",
                                          replace_history, abort_key );
      if( len < 0 ) { replace = false; return 0; }
      if( len > 0 ) replace_string = replace_history.back();
      }
    }
  if( toggle ) backward = !backward;

  Point p = curwin().pointer();			// search point
  if( block )
    {
    if( no_block() || no_block_in_this_file( curwin().buffer() ) ) return 0;
    if( !again || !Block::in_block_or_end( curwin().buffer(), p ) )
      { if( backward ) p = Block::end(); else p = Block::begin(); }
    }

  const Buffer * prev_replace = 0;
  const char * msg = 0;
  bool found = false, ask = true, found_any = false, wrapped = false;
  bool empty_replace = false, feedback = true;
  const int handles = global ? Bufhandle_vector::handles() : 1;
  for( int i = handles - 1; i >= 0 && !found; --i )
    {
    if( feedback )
      { feedback = false;
        Screen::show_feedback( ask ? "searching..." : "replacing..." ); }
    Window & w = data[curwin_];
    std::vector< std::string > regex_pieces;
    if( replace && empty_replace && !backward ) w.buffer().pnext( p );
    Point p1 = p, p2 = p;
    { const int tmp = Regex::find( w.buffer(), p1, p2, search_string,
                                   regex_pieces, icase, backward );
    if( tmp == 2 ) { msg = "Search error: check syntax"; break; }
    found = ( tmp == 0 ); }
    empty_replace = ( found && ( p1 == p2 ) );
    if( found && block &&
        ( p1 < Block::begin() || p1 >= Block::end() || p2 > Block::end() ) )
      found = false;
    if( backward ) p = p1; else p = p2;

    if( found )
      {
      if( block && RC::editor_options().rectangle_mode &&
          ( p1.col < Block::begin().col || p2.col > Block::end().col ) )
        { ++i; found = false; continue; }
      found_any = true;
      if( !replace ) { w.update_points( p, true, true ); break; }
      Block::save_block_position(); w.update_points( p, false, true );
      if( ask )
        {
        const bool rm = RC::editor_options().rectangle_mode;
        RC::editor_options().rectangle_mode = false;
        Block::set_block( w.buffer(), p1, p2 ); w.repaint();
        RC::editor_options().rectangle_mode = rm;
        }
      while( true )
        {
        int key = 0;
        if( ask || Screen::kbhit() >= 0 )
          {
          key = Screen::show_message( "Replace (Y)es (N)o (R)est (^C to abort): ", true );
          if( key == 3 ) { if( !ask ) msg = "Interrupted"; break; }
          if( ask )
            {
            if( key == 'n' || key == 'N' ) { found = false; break; }
            if( key == 'r' || key == 'R' ) ask = false;
            }
          }
        if( !ask || key == 'y' || key == 'Y' )
          {
          Block::restore_block_position();
          if( !ask && prev_replace == &w.buffer() ) w.buffer().force_append();
          const bool done =
            Regex::replace( w.buffer(), p1, p2, replace_string, regex_pieces );
          Block::save_block_position();
          if( done )
            {
            if( !ask ) prev_replace = &w.buffer();
            if( empty_replace && ( p1 < p2 ) ) empty_replace = false;
            found = false; if( backward ) p = p1; else p = p2;
            w.update_points( p, false, true );
            }
          else
            {
            prev_replace = 0; w.buffer().reset_appendable();
            if( w.buffer().options.read_only ) msg = "Read only";
            else msg = "Replace error: check syntax";
            }
          break;
          }
        }
      ++i; Block::restore_block_position();
      if( ask || found ) Screen::repaint( &w.buffer() );
      feedback = !block;
      continue;
      }

    prev_replace = 0; w.buffer().reset_appendable();
    if( !ask && ( block || wrapped || !RC::editor_options().search_wrap ) )
      Screen::repaint( &w.buffer() );
    if( block ) break;
    if( !wrapped )
      {
      if( global )
        {
        if( backward ) { prev(); p = curwin().buffer().eof(); }
        else { next(); p = curwin().buffer().bof(); }
        if( i == 0 ) { wrapped = true; ++i; }
        }
      else if( RC::editor_options().search_wrap )
        {
        if( backward ) p = w.buffer().eof();
        else p = w.buffer().bof();
        wrapped = true; ++i;
        }
      }
    }
  data.back().repaint();			// remove feedback
  if( !msg )
    {
    if( !found_any )
      {
      if( block ) msg = "Not found (search restricted to marked block)";
      else msg = "Not found";
      }
    else if( wrapped ) msg = "Wrapped";
    else if( !ask ) msg = "Done";
    }
  return msg;
  }
